# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from multiprocessing import Process
import numpy as np
import collections
from adaptdl.reducer import Reducer
import portpicker
import signal
import faulthandler

root_host = "127.0.0.1"
DEFAULT_REDUCER_PORT = portpicker.pick_unused_port()


def main(rank, size):
    faulthandler.enable(all_threads=True)
    faulthandler.register(signal.SIGUSR1, all_threads=True, chain=False)

    reducer = Reducer(rank, size, root_host, DEFAULT_REDUCER_PORT)

    if rank == 0:
        batch_size = 28
        x = {"foo": 1}
    else:
        x = {"bar": 1}
        batch_size = 0

    # start a async reducer
    ax = reducer.allreduce_async(np.asarray([1, 1, 1]))

    # do a bunch of bcasts
    batch_size = reducer.broadcast(batch_size)
    batch_size = reducer.broadcast(batch_size)
    batch_size = reducer.broadcast(batch_size)
    assert batch_size == 28

    # do allreduce on Counter
    x = reducer.allreduce(collections.Counter(x))
    assert x["foo"] == 1
    assert x["bar"] == size - 1

    # collect the allreduce_async result
    ax = ax.result()
    assert np.allclose(ax, size * np.asarray([1, 1, 1]))

    # try to simulate a training loop
    x = None
    for _ in range(10):
        if x:
            x = x.result()
            assert np.allclose(x, size * np.asarray([1, 1, 1]))
        x = reducer.allreduce_async(np.asarray([1, 1, 1]))


def test_reducer():
    size = 3  # number of replicas
    processes = []
    for rank in range(size):
        p = Process(target=main, args=(rank, size), daemon=True)
        p.start()
        processes.append(p)

    for p in processes[1:]:
        p.join()

    processes[0].join()

    # check exceptions raised by the processes
    for p in processes:
        assert not p.exitcode



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import logging
import requests
from collections import OrderedDict
from adaptdl.goodput import PerfParams
import adaptdl.env
from types import MappingProxyType


LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


# make immutable proxies of globals
PERF_PARAMS = MappingProxyType(
    OrderedDict({k: 0.0 for k in PerfParams._fields}))

SCHED_HINTS = MappingProxyType({'initBatchSize': 0,
                                'localBszBounds': None,  # [min, max]
                                'globalBatchSize': None,
                                'maxBatchSize': 0,
                                'maxProfiledReplicas': 0,
                                'gradientAccumulation': False,
                                'gradParams': None,
                                'perfParams': None})


def post_sched_hints(sched_hints, job_key):
    url = adaptdl.env.supervisor_url()
    if not url or url == "":
        return  # skip
    headers = {"Content-Type": "application/json"}
    try:
        for k in sched_hints:
            assert k in SCHED_HINTS  # validate

        response = \
            requests.put(url=f"{url}/hints/{job_key}",
                         data=json.dumps(sched_hints),
                         headers=headers)
        if response.status_code != 200:
            LOG.warning(f"Received {response.status_code}")
    except Exception as e:
        LOG.warning(f"{e}")



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
if "darwin" in sys.platform.lower():
    # To avoid multiple runs of the model code
    # https://pythonspeed.com/articles/python-multiprocessing/
    import multiprocessing
    multiprocessing.set_start_method('fork')

import logging
import portpicker
import requests
import torch.distributed
import pkg_resources

import adaptdl.collective
import adaptdl.env
import semver
from .epoch import current_epoch, finished_epochs, remaining_epochs_until
from .data import current_dataloader, AdaptiveDataLoader, ElasticSampler
from .parallel import AdaptiveDataParallel
from .accumulator import Accumulator

logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


def version_check(version):
    if semver.VersionInfo.isvalid(version) and \
            version != "0.0.0":
        return True
    else:
        return False


def init_process_group(backend):
    url = adaptdl.env.supervisor_url()
    if url:
        key = adaptdl.env.job_id()
        group = adaptdl.env.num_restarts()
        while True:
            response = requests.get(url=f"{url}/discover/{key}/{group}")
            if response.status_code != 408:  # Timeout.
                break
        response.raise_for_status()
        master_addr = response.json()[0]
        sched_version = adaptdl.env.adaptdl_sched_version()
        trainer_version = pkg_resources.get_distribution("adaptdl").version
        if version_check(sched_version) and version_check(trainer_version):
            trainer_ver_maj = semver.VersionInfo.parse(trainer_version).major
            sched_ver_maj = semver.VersionInfo.parse(sched_version).major
            if trainer_ver_maj != sched_ver_maj:
                raise Exception('adaptdl version {} is incompatible with'
                                'scheduler version {}'.format(trainer_version,
                                                              sched_version))
    else:
        master_addr = adaptdl.env.master_addr()

    master_port = adaptdl.env.master_port()

    # Initialize collective module.
    adaptdl.collective.initialize(master_addr, master_port)

    # Initialize torch.distributed.
    torch_port = adaptdl.collective.broadcast(portpicker.pick_unused_port())
    init_method = "tcp://{}:{}?rank={}&world_size={}".format(
            master_addr, torch_port, adaptdl.env.replica_rank(),
            adaptdl.env.num_replicas())
    LOG.info("Initializing torch.distributed using %s", init_method)
    torch.distributed.init_process_group(backend, init_method)

    LOG.info("torch.distributed initialized")


__all__ = [
    "init_process_group",
    "current_epoch",
    "finished_epochs",
    "remaining_epochs_until",
    "current_dataloader",
    "AdaptiveDataLoader",
    "ElasticSampler",
    "AdaptiveDataParallel",
    "Accumulator",
]



from adaptdl.conftest import elastic_multiprocessing


@elastic_multiprocessing
def test_allreduce():
    import adaptdl.collective
    import adaptdl.env
    adaptdl.collective.initialize("0.0.0.0")
    # allreduce with default reduce_fn (addition).
    result = adaptdl.collective.allreduce(adaptdl.env.replica_rank())
    assert result == sum(range(adaptdl.env.num_replicas()))
    # allreduce with custom reduce_fn (set union).
    result = adaptdl.collective.allreduce(
        {adaptdl.env.replica_rank()},
        reduce_fn=lambda a, b: a | b)
    assert result == set(range(adaptdl.env.num_replicas()))
    return [5, 0][adaptdl.env.num_restarts()]


@elastic_multiprocessing
def test_allreduce_async():
    import adaptdl.collective
    import adaptdl.env
    adaptdl.collective.initialize("0.0.0.0")
    future_1 = adaptdl.collective.allreduce_async(1)
    future_2 = adaptdl.collective.allreduce_async(2)
    future_3 = adaptdl.collective.allreduce_async(3)
    assert future_2.result() == 2 * adaptdl.env.num_replicas()
    assert future_1.result() == 1 * adaptdl.env.num_replicas()
    assert future_3.result() == 3 * adaptdl.env.num_replicas()
    return [5, 0][adaptdl.env.num_restarts()]


@elastic_multiprocessing
def test_broadcast():
    import adaptdl.collective
    import adaptdl.env
    adaptdl.collective.initialize("0.0.0.0")
    result = adaptdl.collective.broadcast(adaptdl.env.replica_rank())
    assert result == 0
    return [5, 0][adaptdl.env.num_restarts()]



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import signal


logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


def get_exit_flag():
    return EXIT_FLAG


def _handler(signum, frame):
    global EXIT_FLAG
    EXIT_FLAG = True
    LOG.debug("Got signal {}...".format(signum))
    if signum == signal.SIGINT:
        LOG.info("Got SIGINT, exiting gracefully... "
                 "Send signal again to force exit.")
        signal.signal(signal.SIGINT, SIGINT_HANDLER)


EXIT_FLAG = False
SIGINT_HANDLER = signal.getsignal(signal.SIGINT)
signal.signal(signal.SIGTERM, _handler)
signal.signal(signal.SIGINT, _handler)



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from adaptdl.conftest import elastic_multiprocessing


@elastic_multiprocessing
def test_duplicate():
    from adaptdl.env import num_restarts
    from adaptdl.checkpoint import State
    state_1 = State("state_1")  # noqa: F841
    state_2 = State("state_2")  # noqa: F841
    with pytest.raises(ValueError):
        state_dup = State("state_1")  # noqa: F841
    return [2, 0][num_restarts()]


@elastic_multiprocessing
def test_save_load():
    import pickle
    from adaptdl.checkpoint import State, save_all_states, load_state
    from adaptdl.env import replica_rank, num_restarts

    class TestState(State):
        def __init__(self, name):
            super().__init__(name)
            self.synced = False

        def sync(self):
            self.synced = True

        def save(self, fileobj):
            assert replica_rank() == 0  # Should only be called from rank 0.
            pickle.dump(self.value, fileobj)

        def load(self, fileobj):
            # Should load the correct value.
            self.value = pickle.load(fileobj)

    state_1 = TestState("state_1")
    state_2 = TestState("state_2")

    if num_restarts() == 0:
        # Save all state.
        state_1.value = 10
        state_2.value = 20
        save_all_states()
        assert state_1.synced and state_2.synced
        return 2  # Restart with 2 replicas.
    elif num_restarts() == 1:
        load_state(state_1)
        load_state(state_2)
        assert state_1.value == 10
        assert state_2.value == 20
    else:
        assert False



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections
import pickle
import time

import numpy as np

import adaptdl.checkpoint
import adaptdl.collective
import adaptdl.env
from adaptdl.goodput import GoodputFunction, fit_perf_params
from adaptdl.sched_hints import SCHED_HINTS, PERF_PARAMS, post_sched_hints


def profile_step_start(atomic_bsz):
    state = _metrics_state()
    state.atomic_bsz = atomic_bsz
    state.step_start = time.time()
    state.sync_time = 0.0


def profile_sync_time(sync_time):
    _metrics_state().sync_time += sync_time


_PREV_REPORT = None


def profile_step_commit(accumulation_step=False):
    global _PREV_REPORT
    state = _metrics_state()
    step_time = time.time() - state.step_start
    num_nodes = adaptdl.env.num_nodes()
    num_replicas = adaptdl.env.num_replicas()
    key = (num_nodes, num_replicas, state.atomic_bsz)
    if accumulation_step:
        state.profile[key]["accum_step_time"] += step_time
        state.profile[key]["accum_count"] += 1
    else:
        state.profile[key]["optim_step_time"] += step_time
        state.profile[key]["optim_sync_time"] += state.sync_time
        state.profile[key]["optim_count"] += 1
    del state.atomic_bsz
    del state.step_start
    del state.sync_time
    if not accumulation_step:
        if _PREV_REPORT is None:
            _PREV_REPORT = time.time()
        if adaptdl.env.replica_rank() == 0 and time.time() - _PREV_REPORT > 30:
            _fit_perf_params()
            _report_sched_hints()
            _PREV_REPORT = time.time()


_GRAD_PARAM_DICT = {}


def update_grad_params(edp_key, grad_norm_sqr, grad_variance):
    global _GRAD_PARAM_DICT
    _GRAD_PARAM_DICT[edp_key] = np.asarray([grad_norm_sqr, grad_variance])
    grad_params = sum(_GRAD_PARAM_DICT.values())
    _metrics_state().grad_params = (grad_params[0], grad_params[1])


def update_progress(progress):
    _metrics_state().progress = progress


def get_progress():
    return _metrics_state().progress


def set_batch_size(init_batch_size, max_batch_size, local_bsz_bounds,
                   gradient_accumulation):
    state = _metrics_state()
    state.init_batch_size = init_batch_size
    state.max_batch_size = max_batch_size
    state.local_bsz_bounds = local_bsz_bounds
    state.gradient_accumulation = gradient_accumulation


def get_goodput_fn(metrics_options="default"):
    state = _metrics_state()
    if state.grad_params is None or state.perf_params is None:
        return None
    return GoodputFunction(state.perf_params, state.grad_params,
                           state.init_batch_size, metrics_options)


def _fit_perf_params():
    state = _metrics_state()
    profile = {k: v for k, v in state.profile.items() if v.get("optim_count")}
    # Convert profile into numpy arrays.
    num_nodes, num_replicas, atomic_bsz = (
        np.array(k) for k in zip(*profile.keys()))
    accum_step_time = np.array([v.get("accum_step_time", 0.0)
                                for v in profile.values()])
    accum_count = np.array([v.get("accum_count", 0) for v in profile.values()])
    optim_step_time = np.array([v.get("optim_step_time", 0.0)
                                for v in profile.values()])
    optim_sync_time = np.array([v.get("optim_sync_time", 0.0)
                                for v in profile.values()])
    optim_count = np.array([v.get("optim_count", 0) for v in profile.values()])
    assert np.all(optim_count > 0)
    # Non-sync time during optimization steps should be approximately equal to
    # accumulation step time, combine those data points.
    assert np.all(optim_step_time >= optim_sync_time)
    accum_step_time += optim_step_time - optim_sync_time
    accum_count += optim_count
    accum_step_time /= accum_count
    optim_step_time /= optim_count
    state.perf_params = fit_perf_params(num_nodes, num_replicas, atomic_bsz,
                                        accum_step_time, optim_step_time)


def _report_sched_hints():
    assert adaptdl.env.replica_rank() == 0
    state = _metrics_state()
    # Scheduling hints
    sched_hints = SCHED_HINTS.copy()
    sched_hints["perfParams"] = {k: v for (k, v) in
                                 zip(PERF_PARAMS.keys(),
                                 state.perf_params)}
    sched_hints["maxBatchSize"] = state.max_batch_size
    sched_hints["localBszBounds"] = state.local_bsz_bounds
    sched_hints["initBatchSize"] = state.init_batch_size
    if state.grad_params:
        sched_hints["gradParams"] = {}
        sched_hints["gradParams"]["norm"] = state.grad_params[0]
        sched_hints["gradParams"]["var"] = state.grad_params[1]
    sched_hints["maxProfiledReplicas"] = max(key[1] for key in state.profile)
    sched_hints["gradientAccumulation"] = state.gradient_accumulation
    post_sched_hints(sched_hints, adaptdl.env.job_id())


class _MetricsState(adaptdl.checkpoint.State):
    def __init__(self):
        super().__init__("adaptdl-metrics")
        self.profile = collections.defaultdict(collections.Counter)
        self.perf_params = None
        self.grad_params = None
        self.init_batch_size = None
        self.max_batch_size = None
        self.local_bsz_bounds = None
        self.gradient_accumulation = False
        self.progress = 0.0  # Progress in scale-invariant iterations.

    def save(self, fileobj):
        pickle.dump(self.profile, fileobj)
        pickle.dump(self.perf_params, fileobj)
        pickle.dump(self.grad_params, fileobj)
        pickle.dump(self.init_batch_size, fileobj)
        pickle.dump(self.max_batch_size, fileobj)
        pickle.dump(self.local_bsz_bounds, fileobj)
        pickle.dump(self.gradient_accumulation, fileobj)
        pickle.dump(self.progress, fileobj)

    def load(self, fileobj):
        self.profile = pickle.load(fileobj)
        self.perf_params = pickle.load(fileobj)
        self.grad_params = pickle.load(fileobj)
        self.init_batch_size = pickle.load(fileobj)
        self.max_batch_size = pickle.load(fileobj)
        self.local_bsz_bounds = pickle.load(fileobj)
        self.gradient_accumulation = pickle.load(fileobj)
        self.progress = pickle.load(fileobj)


def _metrics_state():
    global _METRICS_STATE
    if _METRICS_STATE is None:
        _METRICS_STATE = _MetricsState()
        adaptdl.checkpoint.load_state(_METRICS_STATE)
    return _METRICS_STATE


_METRICS_STATE = None



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import pickle
import socket
import threading
import time
import traceback
import sys


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class Future(object):
    def __init__(self, reducer, key):
        self._reducer = reducer
        self._key = key

    def result(self):
        try:
            return self._result
        except AttributeError:
            while self._key not in self._reducer._result_map:
                try:
                    key, result = pickle.load(self._reducer._sockfile)
                    self._reducer._result_map[key] = result
                except Exception as e:
                    logger.error(f"reducer._rank = {self._reducer._rank}"
                                 f" is exiting unexpectedly because of {e}")
                    raise
            self._result = self._reducer._result_map.pop(self._key)
            return self._result


def default_reduce_fn(a, b):
    a += b
    return a


class Reducer(object):
    """
    Simple asynchronous (all)reduce operations on python objects. Assumes all
    invokations to allreduce, allreduce_async, and Future.result happen in the
    same order across all processes.
    """

    def __init__(self, rank, replicas, root_host, root_port):
        self._root_port = root_port
        self._result_map = {}
        self._next_key = 0
        self._rank = rank

        if rank == 0:
            self._reduce_fn_map = {}
            threading.Thread(target=self._run_server,
                             args=(self._root_port, replicas),
                             daemon=True).start()
        # Keep retrying connection, because (1) the root pod might not have
        # a registered domain name yet, and (2) the root server socket might
        # not be bound yet.
        exception_cnt = 0
        while True:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            if (exception_cnt > 25):
                logger.error("Could not connect to root after max "
                             "retries, exiting...")
                break
            try:
                if (self._root_port == 0):
                    # waiting for server to get a valid port in local mode
                    raise ConnectionRefusedError
                logger.info(f"rank {rank} connecting to {root_host} "
                            f"on port {self._root_port}")
                sock.connect((root_host, self._root_port))
            except ConnectionRefusedError:
                logger.warning("Could not connect to root, trying again...")
                exception_cnt += 1
                time.sleep(5)
            else:
                break
        self._sockfile = sock.makefile("rwb")
        pickle.dump(rank, self._sockfile)
        self._sockfile.flush()

    def broadcast(self, obj):
        """
        Broadcast a value from replica 0 to all other replicas. Currently uses
        allreduce with left-projection.
        """
        return self.allreduce(obj, lambda x, y: x)

    def allreduce(self, obj, reduce_fn=default_reduce_fn):
        future = self.allreduce_async(obj, reduce_fn)
        return future.result()

    def allreduce_async(self, obj, reduce_fn=default_reduce_fn):
        key = self._next_key
        self._next_key += 1
        try:
            self._reduce_fn_map[key] = reduce_fn
        except AttributeError:
            pass
        pickle.dump(obj, self._sockfile)
        # TODO - flush throws an unhandled exception
        self._sockfile.flush()
        return Future(self, key)

    def _run_server(self, port, replicas):
        try:
            listener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            listener.bind(("0.0.0.0", port))
            if port == 0:
                # local mode
                self._root_port = listener.getsockname()[1]
            listener.listen(replicas)
            # wait for connections from all clients
            logger.info(f"Master waiting for connections on {port}")
            clients = [None] * replicas
            while None in clients:
                client = listener.accept()[0].makefile("rwb")
                rank = pickle.load(client)
                assert clients[rank] is None
                clients[rank] = client
            # main server loop
            key = 0
            while True:
                for rank, client in enumerate(clients):
                    obj = pickle.load(client)
                    if rank == 0:
                        result = obj
                        reduce_fn = self._reduce_fn_map.pop(key)
                    else:
                        result = reduce_fn(result, obj)
                # Respond to clients in reverse order, with rank 0 last.
                # Prevents deadlocks where the rank 0 client gets unblocked
                # first and grabs the GIL in a later operation, blocking this
                # server from responding to the remaining replicas.
                for client in reversed(clients):
                    pickle.dump((key, result), client)
                    client.flush()
                key += 1
        except Exception:
            traceback.print_exception(*sys.exc_info())
            exit(1)



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from adaptdl.goodput import GoodputFunction, PerfParams, GradParams
import itertools
import numpy as np
import pytest

RNG = np.random.RandomState(0)
PERF_PARAMS = [PerfParams(*RNG.gamma(2.0, 2.0, [7])) for i in range(10)]
GRAD_PARAMS = [GradParams(*RNG.gamma(2.0, 2.0, [2])) for i in range(10)]


def groupby_indices(*args):
    _, indices = np.unique(np.stack(args), axis=1, return_inverse=True)
    groups = {}
    for i, g in enumerate(indices):
        groups.setdefault(g, []).append(i)
    return list(groups.values())


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_evaluate(perf_params, grad_params):
    init_batch_size = 16
    goodput_fn = GoodputFunction(perf_params, grad_params, init_batch_size)
    # Generate a range of different goodput function arguments.
    num_nodes = np.array([1, 2, 3, 4])
    num_replicas = np.array([1, 2, 4, 8])
    atomic_bsz = np.array([8, 12, 16, 20, 24])
    accum_steps = np.array([0, 1, 2, 3, 4])
    # Cartesian product.
    num_nodes, num_replicas, atomic_bsz, accum_steps = \
        map(np.array, zip(*itertools.product(num_nodes, num_replicas,
                                             atomic_bsz, accum_steps)))
    # Only keep valid arguments.
    valid = np.logical_and(num_nodes <= num_replicas, init_batch_size
                           <= num_replicas * atomic_bsz * accum_steps)
    num_nodes = num_nodes[valid]
    num_replicas = num_replicas[valid]
    atomic_bsz = atomic_bsz[valid]
    accum_steps = accum_steps[valid]
    # Evaluate goodput.
    goodput = goodput_fn(num_nodes, num_replicas, atomic_bsz, accum_steps)
    throughput = goodput_fn.throughput(num_nodes, num_replicas,
                                       atomic_bsz, accum_steps)
    efficiency = goodput_fn.efficiency(num_replicas * atomic_bsz
                                       * (accum_steps + 1))
    # Check basic invariants.
    assert np.all(0 <= throughput)
    assert np.all(0 <= efficiency) and np.all(efficiency <= 1)
    assert np.allclose(goodput, throughput * efficiency)
    # Increasing batch size should decrease efficiency.
    batch_size = num_replicas * atomic_bsz * (accum_steps + 1)
    sort = np.argsort(batch_size)
    assert np.all(np.diff(efficiency[sort]) <= 0)
    # All else equal, increasing atomic_bsz should increase throughput.
    for indices in groupby_indices(num_nodes, num_replicas, accum_steps):
        sort = np.argsort(atomic_bsz[indices])
        assert np.all(np.diff(throughput[indices][sort]) >= 0)
        # Increasing throughput should experience diminishing returns.
        if len(indices) > 1:
            diffx = np.diff(atomic_bsz[indices][sort])
            diffy = np.diff(throughput[indices][sort])
            assert np.all(diffx[:-1] * diffy[1:] - diffx[1:] * diffy[:-1] <= 0)
    # All else equal, scalability is sublinear with respect to num_replicas.
    for indices in groupby_indices(num_nodes, atomic_bsz, accum_steps):
        scalability = throughput / num_replicas
        sort = np.argsort(num_replicas[indices])
        assert np.all(np.diff(scalability[indices][sort]) <= 0)


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_optimize_no_bounds(perf_params, grad_params):
    goodput_fn = GoodputFunction(perf_params, grad_params, 128)
    goodput, bsz, steps = goodput_fn.optimize(1, 3)
    assert(bsz == 128//3 + 1), "expected bsz = 43, got {}".format(bsz)
    assert(isinstance(goodput, float))

    replicas = np.asarray([1, 2, 3, 4, 5])
    # single-node
    goodput, bsz, steps = goodput_fn.optimize(np.ones_like(replicas), replicas)
    assert(bsz.shape == (5,))
    assert(np.all(bsz == np.ceil(128 / replicas).astype(int)))
    assert(goodput.shape == (5,))
    assert(bsz[0] == 128)
    assert(np.all(steps == 0))
    # multi-node
    goodput, bsz, steps = goodput_fn.optimize(replicas, replicas)
    assert(bsz.shape == (5,))
    assert(np.all(bsz == np.ceil(128 / replicas).astype(int)))
    assert(goodput.shape == (5,))
    assert(bsz[0] == 128)
    assert(np.all(steps == 0))


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_optimize_local_bounds(perf_params, grad_params):
    fun = GoodputFunction(perf_params, grad_params, 128)
    goodput, bsz, steps = fun.optimize(1, 1, atomic_bsz_range=(64, 256))
    assert(bsz == 128), "expected bsz = 128, got {}".format(bsz)
    assert(isinstance(goodput, float))

    replicas = np.asarray(range(1, 100))
    # single-node
    goodput, bsz, steps = fun.optimize(np.ones_like(replicas), replicas,
                                       atomic_bsz_range=(64, 256))
    assert(np.all(bsz >= np.ceil(128 / replicas).astype(int)))
    assert(np.all(np.logical_or(bsz >= (64), goodput == 0.0)))
    assert(np.all(bsz <= (256)))
    assert(np.all(bsz * replicas <= 100 * 128))
    assert(bsz[0] == 128)
    assert(np.all(steps == 0))
    # multi-node
    goodput, bsz, steps = fun.optimize(replicas, replicas,
                                       atomic_bsz_range=(64, 256))
    assert(np.all(bsz >= np.ceil(128 / replicas).astype(int)))
    assert(np.all(np.logical_or(bsz >= (64), goodput == 0.0)))
    assert(np.all(bsz <= (256)))
    assert(np.all(bsz * replicas <= 100 * 128))
    assert(bsz[0] == 128)
    assert(np.all(steps == 0))


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_optimize_max_bounds(perf_params, grad_params):
    fun = GoodputFunction(perf_params, grad_params, 128)
    goodput, bsz, steps = fun.optimize(1, 1, max_batch_size=1280)
    assert(bsz == 128), "expected bsz = 128, got {}".format(bsz)
    assert(isinstance(goodput, float))

    replicas = np.asarray(range(1, 100))
    # single-node
    goodput, bsz, steps = fun.optimize(np.ones_like(replicas), replicas,
                                       max_batch_size=1280)
    assert(np.all(bsz >= np.ceil(128 / replicas).astype(int)))
    assert(np.all(bsz * replicas <= 1280 + replicas))
    assert(bsz[0] == 128)
    assert(np.all(steps == 0))
    # multi-node
    goodput, bsz, steps = fun.optimize(replicas, replicas, max_batch_size=1280)
    assert(np.all(bsz >= np.ceil(128 / replicas).astype(int)))
    assert(np.all(bsz * replicas <= 1280 + replicas))
    assert(bsz[0] == 128)
    assert(np.all(steps == 0))


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_optimize_all_bounds(perf_params, grad_params):
    fun = GoodputFunction(perf_params, grad_params, 128)
    goodput, bsz, steps = fun.optimize(1, 1, max_batch_size=1280,
                                       atomic_bsz_range=(64, 256))
    assert(bsz == 128), "expected bsz = 128, got {}".format(bsz)
    assert(isinstance(goodput, float))

    replicas = np.asarray(range(1, 20))
    # single-node
    goodput, bsz, steps = fun.optimize(np.ones_like(replicas), replicas,
                                       max_batch_size=1280,
                                       atomic_bsz_range=(64, 256))
    assert(np.all(np.logical_or(bsz >= np.ceil(128 / replicas).astype(int),
                                goodput == 0.0)))
    assert(np.all(np.logical_or(bsz >= (64),
                                goodput == 0.0)))
    assert(np.all(bsz <= (256)))
    assert(np.all(np.logical_or(bsz * replicas <= 1280 + replicas,
                                goodput == 0.0)))
    assert(bsz[0] == 128)
    assert(np.all(steps == 0))
    # multi-node
    goodput, bsz, steps = fun.optimize(replicas, replicas,
                                       max_batch_size=1280,
                                       atomic_bsz_range=(64, 256))
    assert(np.all(np.logical_or(bsz >= np.ceil(128 / replicas).astype(int),
                                goodput == 0.0)))
    assert(np.all(np.logical_or(bsz >= (64),
                                goodput == 0.0)))
    assert(np.all(bsz <= (256)))
    assert(np.all(np.logical_or(bsz * replicas <= 1280 + replicas,
                                goodput == 0.0)))
    assert(bsz[0] == 128)
    assert(np.all(steps == 0))
    # multi-node edge case
    replicas = 4
    goodput, bsz, steps = fun.optimize(4, 4, max_batch_size=1024,
                                       atomic_bsz_range=(128, 128))
    assert goodput > 0.0
    assert bsz == 128
    assert steps == 0


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_optimize_accumulation(perf_params, grad_params):
    fun = GoodputFunction(perf_params, grad_params, 128)
    goodput, bsz, steps = fun.optimize(1, 1, max_batch_size=1280,
                                       atomic_bsz_range=(64, 256),
                                       accumulation=True)
    assert(isinstance(goodput, float))

    replicas = np.asarray(range(1, 20))
    # single-node
    goodput, bsz, steps = fun.optimize(np.ones_like(replicas), replicas,
                                       max_batch_size=1280,
                                       atomic_bsz_range=(64, 256),
                                       accumulation=True)
    assert(np.all(np.logical_or(bsz >= np.ceil(128 / replicas).astype(int),
                                goodput == 0.0)))
    assert(np.all(np.logical_or(bsz >= (64),
                                goodput == 0.0)))
    assert(np.all(bsz <= (256)))
    assert(np.all(np.logical_or(bsz * replicas * (steps + 1) <
                                1280 + replicas * (steps + 1),
                                goodput == 0.0)))
    assert(np.all(steps <= 15))
    assert(np.all(steps >= 0))
    assert(np.all(np.logical_or(replicas > 1,
                                np.logical_or(bsz == 128, steps > 0))))
    # multi-node
    goodput, bsz, steps = fun.optimize(replicas, replicas,
                                       max_batch_size=1280,
                                       atomic_bsz_range=(64, 256),
                                       accumulation=True)
    assert(np.all(np.logical_or(bsz >= np.ceil(128 / replicas).astype(int),
                                goodput == 0.0)))
    assert(np.all(np.logical_or(bsz >= (64),
                                goodput == 0.0)))
    assert(np.all(bsz <= (256)))
    assert(np.all(np.logical_or(bsz * replicas * (steps + 1) <
                                1280 + replicas * (steps + 1),
                                goodput == 0.0)))
    assert(np.all(steps <= 15))
    assert(np.all(steps >= 0))
    assert(np.all(np.logical_or(np.multiply(steps, bsz) >= 256,
                                steps == 0)))


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_one_replica_accumulation(perf_params, grad_params):
    fun = GoodputFunction(perf_params, grad_params, 128)

    replicas = np.asarray([1])
    max_batch_sizes = np.asarray(range(128, 128 * 20, 128))
    # single-node
    for max_batch_size in max_batch_sizes:
        goodput, bsz, steps = fun.optimize(np.ones_like(replicas), replicas,
                                           max_batch_size=1280,
                                           atomic_bsz_range=(64, 256),
                                           accumulation=True)
        assert(np.all(np.logical_or(bsz >= (64),
                                    goodput == 0.0)))
        assert(np.all(bsz <= (256)))
        assert(np.all(np.logical_or(bsz * (steps + 1) <=
                                    max_batch_size,
                                    goodput == 0.0)))
        assert(np.all(np.logical_or(bsz >= np.ceil(128 / replicas).astype(int),
                                    goodput == 0.0)))
        assert(np.all(np.logical_or(bsz * (steps + 1) != 128,
                                    steps == 0)))



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
This module contains simple collective communications primitives which operate
on arbitrary python objects. It is meant to be general but *non-performant*.
Only use these primitives if you are synchronizing small objects which can be
efficiently pickled and operated on. For larger objects, use framework-specific
functions, such as those provided by `torch.distributed`.

The functions in this module should be invoked *in the same order* across all
replicas in the current job. Otherwise, their behavior is undefined and you may
encounter unexpected bugs and errors.
"""

# TODO: Merge the reducer into this module once the previous trainer APIs
# are removed.

import adaptdl.env
from .reducer import Reducer, default_reduce_fn

_REDUCER = None


def initialize(master_addr=None, master_port=None):
    """
    Initialize this module, must be invoked before calling any other functions.
    This function will block until it has been invoked from all replicas.

    Arguments:
        master_addr: address of the replica with rank 0.
        master_port: free port of the replica with rank 0.

    Raises:
        RuntimeError: If this module had already been initialized.
    """
    global _REDUCER
    if _REDUCER is not None:
        raise RuntimeError("{} is already initialized".format(__name__))
    if master_addr is None:
        master_addr = adaptdl.env.master_addr()
    if master_port is None:
        master_port = adaptdl.env.master_port()
    _REDUCER = Reducer(adaptdl.env.replica_rank(),
                       adaptdl.env.num_replicas(),
                       master_addr, master_port)


def teardown():
    """
    Teardown this module, will block until this function has been invoked from
    all replicas.

    Raises:
        RuntimeError: If this module has not been initialized.
    """
    raise NotImplementedError  # TODO


def allreduce(value, reduce_fn=default_reduce_fn):
    """
    Reduces a value across all replicas in such a way that they all get the
    final result. Blocks until this function is invoked by all replicas.

    Arguments:
        value (object): The object which will be reduced together with all
            other replicas.
        reduce_fn (Function): A reduction function which two objects as
            arguments, and returns the resulting reduced object.

    Returns:
        object: Resulting value after being reduced across all replicas.

    Raises:
        RuntimeError: If this module has not been initialized.
    """
    if _REDUCER is None:
        raise RuntimeError("{} has not been initialized".format(__name__))
    return _REDUCER.allreduce(value, reduce_fn)


def allreduce_async(value, reduce_fn=default_reduce_fn):
    """
    Asynchronous version of the `allreduce` function. Does not block, instead
    returns a future which can be used to obtain the result later.

    Arguments:
        value (object): The object which will be reduced together with all
            other replicas.
        reduce_fn (Function): A reduction function which two objects as
            arguments, and returns the resulting reduced object.

    Returns:
        Future: Object from which the result can be obtained later.

    Raises:
        RuntimeError: If this module has not been initialized.
    """
    if _REDUCER is None:
        raise RuntimeError("{} has not been initialized".format(__name__))
    return _REDUCER.allreduce_async(value, reduce_fn)


def broadcast(value):
    """
    Broadcasts a value from the replica of rank 0 to all replicas. Blocks until
    this function is invoked by all replicas.

    Arguments:
        value (object): The object which will be broadcasted from replica 0.
            Ignored on all other replicas.

    Returns:
        object: The value broadcasted from replica 0.

    Raises:
        RuntimeError: If this module has not been initialized.
    """
    if _REDUCER is None:
        raise RuntimeError("{} has not been initialized".format(__name__))
    return _REDUCER.broadcast(value)



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import adaptdl.goodput as goodput
import numpy as np


def test_fit_1():
    # Tests goodput.fit's ability to fit to data generated
    # by its own model class with arbitrary parameters, without
    # gradient accumulation. Serves as a sanity check
    # that the goodput.model fitting works in the most
    # optimistic case.
    size = (1000,)
    nodes = np.random.randint(low=1, high=11, size=size)
    replicas = np.random.randint(low=1, high=nodes+1, size=size)
    local_bsz = np.random.randint(32, 1024, size=size)
    params = goodput.PerfParams(0.1, 0.01, 0.5, 1.0, 1e-6, 1e-6, 1.2)
    accum_step_time = (
        goodput._predict_accum_time(params, local_bsz) +
        np.maximum(np.random.normal(0, 0.001, size=size), 0.0))
    network_time = (
        goodput._predict_network_time(params, nodes, replicas) +
        np.maximum(np.random.normal(0, 0.001, size=size), 0.0))
    gamma = params.gamma
    optim_step_time = (accum_step_time ** gamma
                       + network_time ** gamma) ** (1 / gamma)
    result = goodput.fit_perf_params(
        nodes, replicas, local_bsz,
        accum_step_time, optim_step_time)
    loss_result = goodput._obj_fn(
        result, nodes, replicas, local_bsz,
        accum_step_time, optim_step_time)
    loss_true = goodput._obj_fn(
        params, nodes, replicas, local_bsz,
        accum_step_time, optim_step_time)
    assert(abs(loss_result - loss_true) < 0.1 * loss_true
           or loss_result < loss_true), \
        ("goodput.fit failed to fit model from data generated by",
         "goodput.PerfParams(0.1, 0.01, 0.5, 1.0, 1e-6, 1e-6, 1.2)",
         "parameters: {}".format(result))


def test_fit_2():
    # Tests goodput.fit's ability to fit to data generated
    # by its own model class with arbitrary parameters, with
    # gradient accumulation. Serves as a sanity check
    # that the goodput.model fitting works in the most
    # optimistic case.
    size = (1000,)
    nodes = np.random.randint(low=1, high=11, size=size)
    replicas = np.random.randint(low=1, high=nodes+1, size=size)
    local_bsz = np.random.randint(32, 1024, size=size)
    params = goodput.PerfParams(0.1, 0.01, 0.5, 1.0, 1e-6, 1e-6, 1.2)
    accum_step_time = goodput._predict_accum_time(params, local_bsz) + \
        np.maximum(np.random.normal(0, 0.01, size=size), 0.0)
    network_time = goodput._predict_network_time(params, nodes, replicas) + \
        np.maximum(np.random.normal(0, 0.01, size=size), 0.0)
    gamma = params.gamma
    optim_step_time = (accum_step_time ** gamma
                       + network_time ** gamma) ** (1 / gamma)
    result = goodput.fit_perf_params(
        nodes, replicas, local_bsz, accum_step_time, optim_step_time)
    loss_result = goodput._obj_fn(
        result, nodes, replicas, local_bsz,
        accum_step_time, optim_step_time)
    loss_true = goodput._obj_fn(
        params, nodes, replicas, local_bsz,
        accum_step_time, optim_step_time)
    assert(abs(loss_result - loss_true) < 0.1 * loss_true
           or loss_result < loss_true), \
        ("goodput.fit failed to fit model from data generated by",
         "goodput.PerfParams(0.1, 0.01, 0.5, 1.0, 1e-6, 1e-6, 1.2)",
         "parameters: {}".format(result))



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from adaptdl.conftest import elastic_multiprocessing
from adaptdl.torch.accumulator import Accumulator


@elastic_multiprocessing
def test_accumulator_restarts():
    import adaptdl.checkpoint
    import adaptdl.collective
    from adaptdl.env import num_restarts, replica_rank
    adaptdl.collective.initialize("0.0.0.0")
    accum = Accumulator()

    if num_restarts() == 0:
        accum["a"] += 15  # Idempotent.
    assert "a" not in accum
    with accum.synchronized():
        assert "a" in accum
        assert accum["a"] == 15
    assert "a" not in accum
    if num_restarts() == 0:
        accum["a"] -= 5  # Idempotent.
        adaptdl.checkpoint.save_all_states()
        return 4  # Restart with 4 replicas.

    if num_restarts() == 1:  # Idempotent.
        accum.update({"a": replica_rank(), "b": replica_rank()})
    assert len(accum) == 0
    with accum.synchronized():
        assert len(accum) == 2
        assert accum["a"] == 16
        assert accum["b"] == 6
    assert len(accum) == 0
    if num_restarts() == 1:
        adaptdl.checkpoint.save_all_states()
        return 2  # Restart with 2 replicas.

    if num_restarts() == 2:  # Idempotent.
        accum -= {"b": 5, "c": 5}
    with accum.synchronized():
        assert accum["a"] == 16
        assert accum["b"] == -4
        assert accum["c"] == -10
        accum.clear()
    with accum.synchronized():
        assert not accum



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from adaptdl.conftest import elastic_multiprocessing


@elastic_multiprocessing
def test_epoch():
    import adaptdl.checkpoint
    from adaptdl.env import num_restarts
    from adaptdl.torch.epoch import (remaining_epochs_until,
                                     current_epoch, finished_epochs)
    total_epochs = 10
    restart_epoch = 5
    assert current_epoch() is None
    if num_restarts() == 0:
        assert finished_epochs() == 0
        expected_epochs = list(range(restart_epoch + 1))
    elif num_restarts() == 1:
        assert finished_epochs() == restart_epoch
        expected_epochs = list(range(restart_epoch, total_epochs))
    else:
        assert False
    for idx, epoch in enumerate(remaining_epochs_until(10)):
        assert epoch == expected_epochs[idx]
        assert current_epoch() == epoch
        assert finished_epochs() == epoch
        if num_restarts() == 0 and epoch == restart_epoch:
            adaptdl.checkpoint.save_all_states()
            return 5  # Restart with 5 replicas.



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
This module provides functionality to Save and load arbitrary state as part of
checkpoint-restart elasticity. The `State` class can be subclassed to define
how to save/load any state to/from persistent storage, so it can be restored
after the current job restarts and resumed from where it left off.
"""

import os
import shutil
import logging

from adaptdl.env import checkpoint_path, replica_rank, num_restarts

logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)

CKPT_DIR_PREFIX = "checkpoint-"

# FIXME: Keeping global state like this will result in memory leaks for
# applications which do not restart too often.
_STATES_TO_NAMES = {}
_NAMES_TO_STATES = {}


class State(object):
    """
    This class implements An arbitrary piece of state which can be saved and
    loaded as part of a checkpoint, and synchronized across all replicas.
    Should be sub-classed to define custom save, load, and sync logic.
    """

    def __init__(self, name):
        """
        Initialize the state object with a unique identifier `name`, which is
        used to refer to the saved object in persistent storage. No two `State`
        objects may share the same `name`.

        Arguments:
            name (str): Unique name of this `State` object.

        Raises:
            ValueError: If a `State` object with the given name already exists.
        """
        if name in _NAMES_TO_STATES:
            raise ValueError("State '{}' already exists".format(name))
        _NAMES_TO_STATES[name] = self
        _STATES_TO_NAMES[self] = name

    def save(self, fileobj):
        """
        This method should be overridden by subclasses to define how the state
        is saved. Is invoked by `save_all_states` and `save_state` to save the
        state into persistent storage.

        Arguments:
            fileobj (BinaryIO): A binary writable file object.
        """
        pass

    def load(self, fileobj):
        """
        This method should be overridden by subclasses to define how the state
        is loaded. Is invoked by `load_state` to load the state from persistent
        storage.

        Arguments:
            fileobj (BinaryIO): A binary readable file object.
        """
        pass

    def sync(self):
        """
        This method should be overridden by subclasses to define how the state
        is synchronized across replicas. This might be necessary to make sure
        the state is consistent before saving it to persistent storage. Is
        invoked by `save_state` before saving the state.
        """
        pass


def _get_tmp_ckpt_dir():
    if checkpoint_path() is None:
        return None

    tmp_dir = os.path.join(checkpoint_path(), "_checkpoint")
    os.makedirs(tmp_dir, exist_ok=True)
    return tmp_dir


def save_all_states():
    """
    Invokes `save_state` on all `State` objects for which `State.skip` is True.
    This function can be used to trigger a global checkpoint and save every
    `State` in the current job.
    """
    for state in _STATES_TO_NAMES:
        save_state(state)

    # Prevent corrupting original state files in case the process got killed
    # during state file writing.
    if replica_rank() == 0 and checkpoint_path() is not None:
        tmp_ckpt_dir = _get_tmp_ckpt_dir()
        ckpt_dir = os.path.join(checkpoint_path(),
                                f"{CKPT_DIR_PREFIX}{num_restarts()}")
        os.rename(tmp_ckpt_dir, ckpt_dir)  # atomic

        for dir_name in os.listdir(checkpoint_path()):
            dir_path = os.path.join(checkpoint_path(), dir_name)
            if dir_name.startswith(CKPT_DIR_PREFIX) and dir_path != ckpt_dir:
                shutil.rmtree(dir_path)


def save_state(state, sync=True):
    """
    Saves a `State` object to persistent storage. First invokes `State.sync` on
    all replicas if `sync` is `True` (default), and then invokes `State.save`
    on the replica of rank 0 only. Note that we save state to a temporary
    folder first. Then, it will be renamed to the formal checkpoint folder
    after all states are saved.

    Arguments:
        state (State): The `State` object to save to persistent storage.
        sync (bool): Whether `State.sync` should be invoked.
    """
    if sync:
        state.sync()

    if replica_rank() == 0 and checkpoint_path() is not None:
        name = _STATES_TO_NAMES[state]
        state_file = os.path.join(_get_tmp_ckpt_dir(), name)

        with open(state_file, "wb") as f:
            state.save(f)


def load_state(state):
    """
    Load the given `State` object from persistent storage. If the object was
    previously saved, then State.load will be invoked with a readable file
    object to load from.

    Arguments:
        state (State): `State` object to load from persistent storage.

    Returns:
        `True` if state was previously saved and `State.load` was invoked,
        `False` otherwise.
    """
    if checkpoint_path() is None:
        return False

    ckpt_dirs = os.listdir(checkpoint_path())
    if not ckpt_dirs:
        LOG.info(f"No checkpoint found in {checkpoint_path()}.")
        return False

    latest_restart_id = 0
    for dir_name in ckpt_dirs:
        if dir_name.startswith(CKPT_DIR_PREFIX):
            restart_id = int(dir_name[len(CKPT_DIR_PREFIX):])
            latest_restart_id = max(latest_restart_id, restart_id)

    if latest_restart_id != num_restarts() - 1:
        LOG.warning("Cannot find checkpoint from the last restart. "
                    f"Loading checkpoint from restart {latest_restart_id}.")

    ckpt_dir = os.path.join(checkpoint_path(),
                            f"{CKPT_DIR_PREFIX}{latest_restart_id}")
    name = _STATES_TO_NAMES[state]
    state_file = os.path.join(ckpt_dir, name)
    if not os.path.isfile(state_file):
        LOG.warning(f"Cannot find state file {state_file}.")
        return False

    with open(state_file, "rb") as f:
        state.load(f)

    return True



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
import logging

from torchtext.data import BPTTIterator
from torchtext.data.dataset import Dataset
from torchtext.data.batch import Batch

import adaptdl.checkpoint
import adaptdl.collective
import adaptdl.env
from adaptdl.torch.data import AdaptiveDataLoaderMixin

logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


class AdaptiveBPTTIterator(BPTTIterator, AdaptiveDataLoaderMixin):
    def __init__(self, dataset, batch_size, bptt_len, **kwargs):
        max_batch_size = kwargs.pop("max_batch_size", None)
        local_bsz_bounds = kwargs.pop("local_bsz_bounds", None)

        BPTTIterator.__init__(self, dataset=dataset, batch_size=batch_size,
                              bptt_len=bptt_len, **kwargs)
        AdaptiveDataLoaderMixin.__init__(self, batch_size)

        self.num_replicas = adaptdl.env.num_replicas()
        self.rank = adaptdl.env.replica_rank()

        if max_batch_size and local_bsz_bounds:
            self._elastic.autoscale_batch_size(max_batch_size,
                                               local_bsz_bounds)

    # The start index changes when there is a rescaling. We recompute a new
    # start index based on how much we covered before the restart
    def _recompute_start(self, prev_curr, prev_end, curr_end):
        if prev_end == 0:
            return prev_curr
        return math.ceil(prev_curr * curr_end / prev_end)

    def __iter__(self):
        with self._elastic.context():
            if self._elastic.skipdone():
                return

            self.batch_size = self._elastic._sync_local_bsz()

            text = self.dataset[0].text
            TEXT = self.dataset.fields['text']
            TEXT.eos_token = None
            text = text + ([TEXT.pad_token] *
                           int(math.ceil(len(text) / self.batch_size) *
                               self.batch_size - len(text)))
            data = TEXT.numericalize(
                [text], device=self.device)
            data = data.view(self.batch_size, -1).t().contiguous()
            dataset = Dataset(examples=self.dataset.examples, fields=[
                ('text', TEXT), ('target', TEXT)])
            end = data.size(0)  # current length of dataset

            # Change in current batch size changes the dimensions of dataset
            # which changes the starting position in the reshaped dataset. The
            # local batch size is also a function of number of replicas, so
            # when we rescale we need to recalculate where to start the
            # iterations from for the new batch size.
            self._elastic.current_index = \
                self._recompute_start(self._elastic.current_index,
                                      self._elastic.end_index, end)
            self._elastic.end_index = end

            # Every replica reads data strided by bptt_len
            start = self._elastic.current_index + (self.bptt_len * self.rank)
            step = self.bptt_len * self.num_replicas

            # The starting index of the highest rank
            highest_start = self._elastic.current_index + \
                (self.bptt_len * (self.num_replicas - 1))

            # Number of steps we will take on the highest rank. We limit
            # iterations on all replicas by this number to prevent asymmetric
            # reduce ops which would result in a deadlock.
            min_steps_in_epoch = max(math.ceil((end - highest_start) / step), 0)  # noqa: E501
            self.iterations = 0
            while True:
                for i in range(start, end, step):
                    self.iterations += 1
                    # Make sure that _elastic.profile is called equal number of
                    # times on all replicas
                    if self.iterations > min_steps_in_epoch:
                        break
                    with self._elastic.profile(self.training and i > 0):
                        seq_len = min(self.bptt_len, data.size(0) - i - 1)
                        assert seq_len > 0
                        batch_text = data[i:i + seq_len]
                        batch_target = data[i + 1:i + 1 + seq_len]
                        if TEXT.batch_first:
                            batch_text = batch_text.t().contiguous()
                            batch_target = batch_target.t().contiguous()
                        yield Batch.fromvars(
                            dataset, self.batch_size,
                            text=batch_text,
                            target=batch_target)
                        self._elastic.current_index += step

                if not self.repeat:
                    break
